------------------------------------------------------------------------------------
-- Company: Univ Lyon, UJM-Saint-Etienne, CNRS, Laboratoire Hubert Curien UMR 5516, F-42023, SAINT-ETIENNE, France
-- Creator: Cédric Marchand  
-- 
-- Create Date:    18:33:09 10/28/2015 
-- Design Name: 
-- Module Name:    control_unit - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity control_unit is
  generic (
    state_size  : in integer := 64;
    key_size    : in integer := 80
  );
  port(
    clk_i     : in std_logic;
    rstb_i    : in std_logic;
    --instruction
    control_i : in std_logic_vector(1 downto 0);
    --output flag
    key_ready : out std_logic;
    busy_o    : out std_logic;
    done_o    : out std_logic;
    --control signal
    T         : in std_logic_vector(7 downto 0);
    ena_t     : out std_logic;
    mode_o    : out std_logic;
    key_sel   : out std_logic;
    data_sel  : out std_logic_vector(1 downto 0)
    );
end control_unit;

architecture Behavioral of control_unit is
-- state nachine states
type SM_type is (
  idle,
  ready,
  load_key,
  load_data_enc,
  load_data_dec,
  release,
  encryption,
  decryption
);

signal c_state : SM_type;
signal n_state : SM_type;

--internal signals
signal key_ready_s  : std_logic;
signal ena_c        : std_logic;
signal ena_r        : std_logic;
--counters
signal load_count   : std_logic_vector(3 downto 0);
signal round_count  : std_logic_vector(1 downto 0);

begin

-- current state
cs_proc : process(rstb_i, clk_i)
begin
  if rstb_i = '0' then
    c_state <= idle;
  elsif rising_edge(clk_i) then
    c_state <= n_state;
  end if;
end process;

--next state
ns_proc : process(c_state, control_i, load_count, T, round_count)
begin
  n_state <= idle;
  case c_state is
  when idle =>
    if control_i = "01" then
      n_state <= load_key;
    end if;
  when load_key =>
    if unsigned(load_count)=(key_size/8)-1 then
      n_state <= ready;
    else
      n_state <= load_key;
    end if;
  when ready =>
    if control_i = "01" then
      n_state <= load_key;
    elsif control_i = "10" then
      n_state <= load_data_enc;
    elsif control_i = "11" then
      n_state <= load_data_dec;
    else
      n_state <= ready;
    end if;
  when load_data_enc =>
    if unsigned(load_count)=(state_size/8)-1 then
      n_state <= encryption;
    else
      n_state <= load_data_enc;
    end if;
  when load_data_dec =>
    if unsigned(load_count)=(state_size/8)-1 then
      n_state <= decryption;
    else
      n_state <= load_data_dec;
    end if;
  when encryption =>
    if T="01111111" and unsigned(round_count)=2 then
      n_state <= release;
    else 
      n_state <= encryption;
    end if;
  when decryption =>
    if T="11111110" and unsigned(round_count)=2 then
      n_state <= release;
    else 
      n_state <= decryption;
    end if;
  when release =>
    if unsigned(load_count)=(state_size/8)-1 then
      n_state <= idle;
    else
      n_state <= release;
    end if;
  when others =>
  end case;
end process;

--output
out_proc : process(c_state)
begin
  key_ready_s <= '0';
  busy_o      <= '0';
  mode_o      <= '0';
  done_o      <= '0';
  ena_c       <= '0';
  ena_r       <= '0';
  key_sel     <= '0';
  data_sel    <= "00";
  case c_state is
  when load_key =>
    busy_o  <= '1';
    ena_c   <= '1';
    key_sel <= '1';
  when ready =>
    key_ready_s <= '1';
  when load_data_enc =>
    busy_o      <= '1';
    key_ready_s <= '1';
    ena_c       <= '1';
    data_sel    <= "01";
  when load_data_dec =>
    busy_o      <= '1';
    key_ready_s <= '1';
    ena_c       <= '1';
    data_sel    <= "01";
  when encryption =>
    busy_o      <= '1';
    key_ready_s <= '1';
    ena_r       <= '1';
    data_sel    <= "10";
  when decryption =>
    busy_o      <= '1';
    key_ready_s <= '1';
    ena_r       <= '1';
    mode_o      <= '1';
    data_sel    <= "11";
  when release =>
    busy_o      <= '1';
    key_ready_s <= '1';
    ena_c       <= '1';
    done_o      <= '1';
    data_sel    <= "01";
  when others =>
  end case;
end process;

key_ready <= key_ready_s;

--load/release counter
L_count : process(rstb_i, clk_i, ena_c)
begin
  if rstb_i = '0' then
    load_count <= (others => '0');
  elsif rising_edge(clk_i) then
    if ena_c = '1' then
      load_count <= std_logic_vector(unsigned(load_count)+1);
    else
      load_count <= (others => '0');
    end if;
  end if;
end process;

--round_count
R_count : process(rstb_i, clk_i, ena_r, round_count)
begin
  if rstb_i = '0' then
    round_count <= (others => '0');
  elsif rising_edge(clk_i) then
    if ena_r = '1' then
      if unsigned(round_count)=2 then
        round_count <= (others => '0');
      else
        round_count <= std_logic_vector(unsigned(round_count)+1);
      end if;
    else
      round_count <= (others => '0');
    end if;
  end if;
end process;

ena_t <= round_count(1);

end Behavioral;

